<?php
/**
 * Ejemplo de uso de SimpleSAMLphp desde PHP
 *
 * Esta aplicación php invoca al SimpleSAMLphp para comprobar su funcionamiento como SP SAML.
 *
 * PHP version >= 8.0
 *
 * LICENSE: Apache License, Version 2.0. http://www.apache.org/licenses/LICENSE-2.0
 *
 * @author     Daniel García <daniel (dot) garcia (at) prise (dot) es>
 * @copyright  2011 Auditoría y Consultoria de Privacidad y Seguridad, S.L.
 * @license    http://www.apache.org/licenses/LICENSE-2.0  Apache License 2.0
 */
use SAML2\XML\saml\NameID;
use SimpleSAML\Auth\Simple;
use SimpleSAML\Auth\State;
use SimpleSAML\Configuration;
use SimpleSAML\Session;

require 'vendor/autoload.php';

ob_start();
$header = null;
$body = null;

try {
    /*
     * Creación objeto SimpleSAMLphp a partir del identificador del SP SAML que se haya creado en
     * el fichero de configuración vendor/simplesamlphp/simplesamlphp/config/authsources.php
     */
    $myURL = getMyUrl();
    $actionParam = filter_input(INPUT_GET, 'action', FILTER_DEFAULT) ?? '';
    $simpleSAMLphpVersion = 'SimpleSAMLphp-'.Configuration::getInstance()->getVersion();
    $spEntityID = 'IdentificadorDeMiServiceProvider';
    $sp = new Simple($spEntityID);

    /*
     * En función de los valores del parámetro HTTP GET 'action' se define el comportamiento de la
     * aplicación.
     *     'login': Inicia el proceso de autenticación.
     *     'forceAuthLogin': Solicita la reautenticación del usuario, independientemente de que
     *     tuviese sesión en este u otro SP.
     *     'logout': Incia el proceso de cierre de sesión.
     *     'loggedout': Comprueba que la sesión se ha cerrado correctamente
     */
    switch($actionParam) {
        case 'login':
            doLogin($sp, $myURL);
            break;
        case 'forceAuthLogin':
            doForceAuthLogin($sp, $myURL);
            break;
        case 'logout':
            doLogout($sp, $myURL);
            break;
        case 'loggedout':
            $body = checkLoggedOut($simpleSAMLphpVersion);
            break;
        default:
            /* Comprueba si el usuario tiene una sesión SAML */
            if (!$sp->isAuthenticated()) {
                Session::getSessionFromRequest()->cleanup();
                $body = getNotLoggedHTML($simpleSAMLphpVersion);
            } else {
                $body = getLoggedHTML($sp, $simpleSAMLphpVersion, $actionParam);
            }
    }
    echo getHTMLHeader();
    echo $body;
} catch (Exception $e) {
    echo getHTMLHeader();
    echo getExceptionHTML($e, $simpleSAMLphpVersion);
}

ob_end_flush();
exit();

/**
 * Invoca función login de SimpleSAMLphp, la cual dispara petición de autenticación SAML
 * Para más detalle sobre parámetros véase documentación de SimpleSAMLphp
 * 
 * @see https://simplesamlphp.org/docs/stable/simplesamlphp-sp-api.html#login
 * @param  Simple $sp
 * @param  string $myUrl
 * @return void
 */
function doLogin(Simple $sp, string $myUrl): void {
    $errorURL = $myUrl.'?action=error';
    $sp->login(
        [
            'ErrorURL' => $errorURL,
            'ReturnTo' => $myUrl,
        ]
        );
    Session::getSessionFromRequest()->cleanup();
}

function getHTMLHeader(): string {;
    $version = 'SimpleSAMLphp-'.Configuration::getInstance()->getVersion();
    return <<<_END_OF_HTML
    <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "DTD/xhtml1-transitional.dtd">
    <html>
        <head>
            <link rel='icon' type='image/png' href='favicon.ico'>
            <!-- Google Fonts -->
            <link rel="stylesheet" href="https://fonts.googleapis.com/css?family=Roboto:300,300italic,700,700italic">
            <!-- CSS Reset -->
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/normalize/8.0.1/normalize.css">
            <!-- Milligram CSS -->
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/milligram/1.4.1/milligram.css">
            <style>
                .wrapLongLine {
                    overflow-wrap: break-word;
                    word-break: break-all;
                    white-space: wrap;
                }
                summary {
                    font-size: 3.6rem;
                    line-height: 1.25;
                    margin-top: 3rem;
                }
            </style>
            <title>Testing {$version}</title>
        </head>
    _END_OF_HTML;
}

/**
 * Invoca función login de SimpleSAMLphp, la cual dispara petición de autenticación SAML.
 * Al pasar el parámetro 'ForceAuthn' => true, implica que se solicita al IdP una
 * reautenticación del usuario, esto es, el IdP debe validar de nuevo las credenciales del
 * usuario.
 * Para más detalle sobre parámetros véase documentación de SimpleSAMLphp
 * 
 * @see https://simplesamlphp.org/docs/stable/simplesamlphp-sp-api.html#login
 * @param  Simple $sp
 * @param  string $myUrl
 * @return void
 */
function doForceAuthLogin(Simple $sp, string $myUrl): void {
    $errorUrl = $myUrl.'?action=error';
    $returnUrl = $myUrl.'?action=forcedAuth';
    $sp->login(
        [
            'ErrorURL' => $errorUrl,
            'ReturnTo' => $returnUrl,
            'ForceAuthn' => true
        ]
    );
}

/**
 * Invoca función logout de SimpelSAMLphp, la cual dispara petición de logout SAML.
 *
 * Se usa el parámetro 'ReturnTo' => $returnUrl, para indicar al IdP que tras realizar el logout
 * debe redirigir el user-agent del usuario a la URL deseada, donde se deberá comprobar
 * que la sesión SAML se ha cerrado correctamente, y se deberá provocar el cierre de sesión de la
 * aplicación.
 *
 * Para más detalle sobre parámetros véase documentación de SimpleSAMLphp
 *
 * @see https://simplesamlphp.org/docs/stable/simplesamlphp-sp-api.html#logout
 * @param  Simple $sp
 * @param  string $myUrl
 * @return void
 */
function doLogout(Simple $sp, string $myUrl): void {
    $returnUrl = $myUrl.'?action=loggedout';
    $sp->logout(
        [
            'ReturnTo' => $returnUrl,
            'ReturnStateParam' => 'LogoutStateID',
            'ReturnStateStage' => 'LogoutStateStage',
        ]
    );
    Session::getSessionFromRequest()->cleanup();
}

/**
 * Comprueba que se haya cerrado correctamente la sesión SAML.
 * Tras realizar la comprobación devuelve contenido HTML para informar al usuario.
 *
 * Importante. Tras comprobar que la sesión se cerro correctamente, se debería invocar el cierre
 * de sesión de la aplicación. En este caso como el script es muy simple no se realiza.
 *
 *
 * @return string
 */
function checkLoggedOut(string $version): string {
    $version = 'SimpleSAMLphp-'.Configuration::getInstance()->getVersion();
    $state = State::loadState((string)$_REQUEST['LogoutStateID'], 'LogoutStateStage');
    $stateDump = var_export($state, true);
    $logoutStatus = $state['saml:sp:LogoutStatus'] ?? ''; /* Only works for SAML SP */
    $statusCode = $logoutStatus['Code'] ?? '';
    if ($statusCode === 'urn:oasis:names:tc:SAML:2.0:status:Success' && !isset($logoutStatus['SubCode'])) {
        $message = "SimpleSAMLphp ha hecho un logout y me ha redirijido aquí";
        // Importante aquí se debería invocar cierre sesiónd de la aplicación
    } else {
        $message = "No se cerrar todas su sesiones. Para estar seguro he ha sido desconectado, necesita cessar su navegador";
    }

    return <<<_END_OF_HTML
        <body><div class="container"><div class="row"><div class="column    ">
            <h1>Proveedor de servicio SAML implementado con {$version}</h1>
            <h2>{$message}</h2>
            <a class="button button-outline" href='./'>Volver</a>
            <hr/>
            <details>
                <summary>Logout State Dump</summary>
                <pre><code>{$stateDump}</code></pre>
            </details>
            </div></div></div></body>
    </html>
_END_OF_HTML;
}

/**
 * Muestra un mensaje indicando que el usuario no tiene sesión y los botones para disparar la
 * petición de autneticación SAML
 *
 * @return string
 */
function getNotLoggedHTML(string $version): string {
    return <<<_END_OF_HTML
    <body><div class="container"><div class="row"><div class="column    ">
        <h1>Proveedor de servicio SAML implementado con {$version}</h1>
        <h2>El usuario no tiene sesión</h2>
        <a class="button button-outline" href='./?action=login'>Login</a>
        <a class="button button-outline" href='./?action=forceAuthLogin'>forceAuth Login</a>
    </div></div></div></body>
</html>
_END_OF_HTML;
}

/**
 * Muestra la información sobre el usuario autenticado, el conjunto de atributos devueltos por
 * el IdP, información sobre el SAML NameID, y un volcado de toda la información relativa a la
 * autneticación.
 *
 * @param  \SimpleSAML\Auth\Simple $sp
 * @param  string                  $version
 * @param  string                  $actionParam
 * @return string
 */
function getLoggedHTML(Simple $sp, string $version, string $actionParam = ''): string {
    $idp = $sp->getAuthData('saml:sp:IdP');
    $nameID = $sp->getAuthData('saml:sp:NameID')->getValue();
    $forcedAuth = getForceAuth($sp, $actionParam);
    $htmlAttributes = getAttributes($sp);
    $htmlNameID = getNameID($sp);
    $htmlAuthnContext = getAuthnContext($sp);
    $htmlAuthData = getAuthDataDump($sp);

    return <<<_END_OF_HTML
    <body><div class="container"><div class="row"><div class="column    ">
        <h1>Proveedor de servicio SAML implementado con {$version}</h1>
        <h2>El usuario SI tiene permisos de acceso</h2>
        <p>Usuario <code>{$nameID}</code> identificado en <code>{$idp}</code></p>
        {$forcedAuth}
        <a class="button button-outline" href='./?action=logout'>Logout</a>
        <a class="button button-outline" href='./?action=forceAuthLogin'>forceAuth Login</a>
        {$htmlAttributes}
        {$htmlNameID}
        {$htmlAuthnContext}
        {$htmlAuthData}
    </div></div></div></body>
</html>
_END_OF_HTML;
}

/**
 * Devuelve una cadena HTML con la información sobre posibles excepciones/errores
 *
 * @param  Exception $e
 * @param  string    $version
 * @return string
 */
function getExceptionHTML(Exception $e, string $version ): string {
    $message = $e->getMessage();
    $trace = print_r($e->getTraceAsString(), true);
    return <<<_END_OF_HTML
    <body><div class="container"><div class="row"><div class="column    ">
        <h1>Proveedor de servicio SAML implementado con {$version}</h1>
        <details>
            <summary>Error en la ejecución</summary>
            <p>{$message}</p>
            <pre>{$trace}</pre>
        </details>
        </div></div></div></body>
</html>
_END_OF_HTML;
}

/**
 * Comprueba si el usuario está autenticado tras una petición de autenticaci´pon ForceAuthn.
 * Mostrando información sobre cúando se produjo la autenticación actual.
 *
 * Esta comprobación es viable gracias al valor del parámetro action que añade en la url de
 * retorno para la petición de autenticación de la función doForceAuthLogin().
 * Si no fuera a sí no sería viable esta comprobación
 *
 * @param  \SimpleSAML\Auth\Simple $sp
 * @param  string                  $actionParam
 * @return string
 */
function getForceAuth(Simple $sp, string $actionParam): string {
    $forcedAuth = '';
    if ($actionParam === 'forcedAuth') {
        $authInstant = (int)$sp->getAuthData('saml:AuthnInstant');
        date_default_timezone_set('Europe/Madrid');
        $validSince = gmdate(DATE_RFC2822, $authInstant);
        $forcedAuth =<<<_END_OF_FORCEAUTH
        <p>
            Volviendo de una <i>Re Autenticación</i> mediante parámetro SAML ForceAuth.<br/>
            Su sesión es valida desde {$validSince}
        </p>
_END_OF_FORCEAUTH;
    }
    return $forcedAuth;
}

/**
 * Devuelve una cadena HTML con el contenido de los atributos del usuario devueltos por el IdP
 *
 * Para más detalle sobre parámetros véase documentación de SimpleSAMLphp
 *
 * @see https://simplesamlphp.org/docs/stable/simplesamlphp-sp-api.html#getattributes
 * @param  \SimpleSAML\Auth\Simple $sp
 * @return string
 */
function getAttributes(Simple $sp): string {
    $attrs = $sp->getAttributes();
    $table = <<<_END_OF_TABLE
        <details>
            <summary>Atributos</summary>
            <table>
                <thead>
                    <tr>
                        <th>Atributo</th>
                        <th>Valore/s</th>
                    </tr>
                </thead>
                <tbody>
_END_OF_TABLE;
    $rows = '';
    foreach ($attrs as $attrName => $attrValues) {
        $htmlValues = getHTMLValues($attrValues);
        $rows .=<<<_END_OF_ROW
                    <tr>
                        <td>{$attrName}</td>
                        <td>
{$htmlValues}
                        </td>
                    </tr>
_END_OF_ROW;
    }
    $table .= $rows;
    $table .=<<<_END_OF_TABLE
                </tbody>
            </table>
        </details>
_END_OF_TABLE;
    return $table;
}

/**
 * Devuelve una cadena HTML con el contenido del elemento <samlp:NameID> de la respuesta de
 * autenticación SAML
 *
 * Para más detalle sobre parámetros véase documentación de SimpleSAMLphp
 *
 * @see https://simplesamlphp.org/docs/stable/simplesamlphp-sp-api.html#getauthdata
 *
 * @param  \SimpleSAML\Auth\Simple $sp
 * @return string
 */
function getNameID(Simple $sp): string {
    /** @var NameID $nameID */
    $nameID = $sp->getAuthData('saml:sp:NameID');
    $nameQualifier = $nameID->getNameQualifier();
    $spNameQualifier = $nameID->getSPNameQualifier();
    $format = $nameID->getFormat();
    $spProvidedID = $nameID->getSPProvidedID();
    $value = $nameID->getValue();

    $htmlValue = '';
    $htmlFormat = '';
    $htmlNameQualifier = '';
    $htmlSPNameQualifier = '';
    $htmlSPProvidedID = '';
    if (!empty($value)) {
        $htmlValue = "<tr><td>Value</td><td><code>{$value}</code></td></tr>";
    }
    if (!empty($format)) {
        $htmlFormat = "<tr><td>Format</td><td><code>{$format}</code></td></tr>";
    }
    if (!empty($nameQualifier)) {
        $htmlNameQualifier = "<tr><td>NameQualifier</td><td><code>{$nameQualifier}</code></td></tr>";
    }
    if (!empty($spNameQualifier)) {
        $htmlSPNameQualifier = "<tr><td>SPNameQualifier</td><td><code>{$spNameQualifier}</code></td></tr>";
    }
    if (!empty($spProvidedID)) {
        $htmlSPProvidedID = "<tr><td>SPProvidedID</td><td><code>{$spProvidedID}</code></td></tr>";
    }

    return <<<_END_OF_TABLE
        <details>
            <summary>SAML NameID</summary>
            <table>
                <thead>
                    <tr><th>Campo</th><th>Valor</th></tr>
                </thead>
                <tbody>
                    {$htmlValue}
                    {$htmlFormat}
                    {$htmlNameQualifier}
                    {$htmlSPNameQualifier}
                    {$htmlSPProvidedID}
                </tbody>
            </table>
        </details>
_END_OF_TABLE;
}

/**
 * Devuelve una cadena HTML con la información sobre el contexto de autneticación devuelto por el
 * IdP
 *
 * Para más detalle sobre parámetros véase documentación de SimpleSAMLphp
 *
 * @see https://simplesamlphp.org/docs/stable/simplesamlphp-sp-api.html#getauthdata
 *
 * @param  \SimpleSAML\Auth\Simple $sp
 * @return string
 */
function getAuthnContext(Simple $sp): string {
    $authnContext = $sp->getAuthData('saml:sp:AuthnContext');
    return <<<_END_OF_HTML
        <details>
            <summary>Authentication Context Dump (saml:sp:AuthnContext)</summary>
            <pre><code>{$authnContext}</code></pre>
        </details>
_END_OF_HTML;
}

/**
 * Devuelve una cadena HTML con el volcado en formato array con los datos disponibles sobre la
 * autenticación SAML
 *
 * Para más detalle sobre parámetros véase documentación de SimpleSAMLphp
 *
 * @see https://simplesamlphp.org/docs/stable/simplesamlphp-sp-api.html#getauthdata
 * 
 * @param  \SimpleSAML\Auth\Simple $sp
 * @return string
 */
function getAuthDataDump(Simple $sp): string {
    $authData = trim(var_export($sp->getAuthDataArray(), true));
    return <<<_END_OF_HTML
        <details>
            <summary>Authentication Data como array PHP</summary>
            <pre><code>{$authData}</code></pre>
        </details>
_END_OF_HTML;
}

/**
 * Genera la URL de la aplicación a partir de la super variable $_SERVER
 *
 * @return string
 */
function getMyUrl(): string {
    $isHTTPS = filter_input(INPUT_SERVER, 'HTTPS', FILTER_VALIDATE_BOOL);
    $urlHost = filter_input(INPUT_SERVER, 'HTTP_HOST', FILTER_DEFAULT);
    $urlPath = filter_input(INPUT_SERVER, 'PHP_SELF', FILTER_DEFAULT, FILTER_NULL_ON_FAILURE) ?? '';
    $urlSchema = $isHTTPS ? 'https://' : 'http://';
    return $urlSchema.$urlHost.$urlPath;
}

/**
 * Devuelve uan cadena HTML para el valor de las celdas de la tabla de atributos
 *
 * @param  array  $values
 * @return string
 */
function getHTMLValues(array $values): string {
    $nValues = count($values);
    if ($nValues === 0) {
        return '';
    } elseif ($nValues === 1){
        if (strlen($values[0]) > 100) {
            return <<<_END_OF_CODE
                        <code class="wrapLongLine">{$values[0]}</code>
_END_OF_CODE;
        }
        return <<<_END_OF_CODE
                        <code >{$values[0]}</code>
_END_OF_CODE;
    } else {
        $output =<<<_END_OF_UL
                        <ul>
_END_OF_UL;
        foreach($values as $aValue) {
            if (strlen($values[0]) > 100) {
                return <<<_END_OF_CODE
                        <code class="wrapLongLine">{$values[0]}</code>
_END_OF_CODE;
            }
            return <<<_END_OF_CODE
                        <code >{$values[0]}</code>
_END_OF_CODE;
        }
        $output .=<<<_END_OF_UL
                        </ul>
_END_OF_UL;
        return $output;
    }
}